/// ===========================================================================
/// C= Parallel Programming Language Support Header
/// ===========================================================================
/// Copyright HOOPOE ProGroup, 2013. All rights reserved
/// ---------------------------------------------------------------------------
///
/// GENERAL NOTES:  C= (pronounced 'see stripes') is a 
///                 Parallel C/C++ Programming Language Extension.
///                 CPX prefix stands for 'C= Parallel eXtension'
///
/// ---------------------------------------------------------------------------
/// LEGAL NOTICE: The terms and conditions of use of this file and the code 
///               herein are subject to applicable laws and an accompanying 
///               license agreement, and may vary depending on the actual 
///               purpose and form of use, including but not limited to
///               research, non-commercial, business purposes, derived 
///               works, as appropriately covered by said license agreement.
///               Customization of keywords provided in a dedicated section 
///               of this file is allowed only to resolve otherwise
///               unavoidable keyword conflicts, while all other modifications
///               are subject to explicit licensing.
/// ---------------------------------------------------------------------------
/// ABI VERSION: 0.3
/// ---------------------------------------------------------------------------

#ifndef _CPX_HEADER_H_
#define _CPX_HEADER_H_

#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0400
#endif

#include <windows.h>
#include <intrin.h>

#ifndef CPX_EXPORTED_API
#define CPX_EXPORTED_API __declspec(dllimport)
#pragma comment(lib, "C=")
#endif

#ifdef __cplusplus
extern "C" {
#endif

WINBASEAPI BOOL WINAPI SwitchToThread(void);

/// ===========================================================================
/// C= Language Customization Part
/// ===========================================================================

/// ---------------------------------------------------------------------------
/// Custom Keywords
/// ---------------------------------------------------------------------------

/// primary operators

#ifndef CPX_MACRO_OMPCOMP

#define parallel                CPX_KEYWORD_PARALLEL

#endif

#define serial                  CPX_KEYWORD_SERIAL

/// language-defined functions (secondary operators)
#define wait                    CPX_KEYWORD_WAIT
#define wake                    CPX_KEYWORD_WAKE
#define pid                     CPX_KEYWORD_PID
#define pix                     CPX_KEYWORD_PIX
#define pno                     CPX_KEYWORD_PNO

/// operators to become implicit with the compiler support
#define barrier                 CPX_KEYWORD_BARRIER
#define coload                  CPX_KEYWORD_COLOAD
#define cpx_sync()

/// operators to merge with C keywords with the compiler support
#define break()                 CPX_KEYWORD_BREAK()
/// try() operator is to be customized by changing its definition (below)

/// modifiers to become implicit with the compiler support
#define shared                  CPX_KEYWORD_SHARED
#define cpx_in                  CPX_KEYWORD_COLOAD_INPUT
#define cpx_out                 CPX_KEYWORD_COLOAD_OUTPUT
#define cpx_inout               CPX_KEYWORD_COLOAD_DUAL
#define cpx_const               CPX_KEYWORD_COLOAD_CONST
#define cpx_range               CPX_KEYWORD_COLOAD_RANGE
#define cpx_error               CPX_KEYWORD_COLOAD_ERROR
#define cpx_cache               CPX_KEYWORD_COLOAD_CACHE
#define cpx_warp                CPX_KEYWORD_COLOAD_WARP

/// ---------------------------------------------------------------------------
/// Custom Macros
/// ---------------------------------------------------------------------------

/// parallel data decomposition macros
#define CPX_RANGE_START         CPX_MACRO_RANGE_START
#define CPX_RANGE_END           CPX_MACRO_RANGE_END

/// ===========================================================================
/// C= Runtime Part
/// ===========================================================================

/// C= runtime functions are prefixed with cpx_exported_
/// C= auxiliary functions are prefixed with cpx_intrinsic_
/// C= data types are prefixed with cpx_datatype_
/// C= global variables are prefixed with cpx_global_
/// C= automatic variables are prefixed with cpx_local_
/// C= enumeration constants are prefixed with cpx_const_
/// C= keyword macros are prefixed with CPX_KEYWORD_
/// C= functional macros are prefixed with CPX_MACRO_
/// C= data macros are prefixed with CPX_CONST_

/// ---------------------------------------------------------------------------
/// Internal Data Structures
/// ---------------------------------------------------------------------------

#pragma pack(push, 1)

typedef struct
{
    size_t stk_bottom;
    size_t stk_diff;
    size_t stk_frame;

    size_t ax;
    size_t bx;
    size_t cx;
    size_t dx;
    size_t si;
    size_t di;
    size_t bp;
    size_t sp;
    size_t ip;

    size_t r8;
    size_t r9;
    size_t r10;
    size_t r11;
    size_t r12;
    size_t r13;
    size_t r14;
    size_t r15;

} cpx_datatype_execution_context;

struct _cpx_datatype_tls_context;

typedef enum {cpx_const_task = 0, cpx_const_thread = 1} cpx_datatype_partype;

typedef struct _cpx_datatype_region_context
{
    /// backward compatibility
    size_t version;

    /// parallel execution type
    volatile cpx_datatype_partype type;

    /// operating context
    int pix;                                    /// current parallel index
    int pno;                                    /// size of parallel region
    volatile long busy;                         /// unsafe-to-return flag
    volatile long lcount;                       /// tasks to launch
    volatile long bcount;                       /// tasks on barrier

    /// context references
    struct _cpx_datatype_tls_context* head;     /// main task's TLS
    struct _cpx_datatype_tls_context* curr;     /// current TLS

    struct _cpx_datatype_region_context* head_rgn;  /// head region

    cpx_datatype_execution_context* idle;       /// idle execution context

    struct _cpx_datatype_region_context* prev_nest;  /// saved tls::rgnctx
    struct _cpx_datatype_region_context* prev_task;  /// saved task list head

    /// register context
    cpx_datatype_execution_context exectx;

} cpx_datatype_region_context;

typedef struct _cpx_datatype_tls_context
{
    /// parallel region context (in-stack)
    cpx_datatype_region_context* rgnctx;

    /// task synchronization objects
    CRITICAL_SECTION mutex;
    HANDLE wakesig;

    /// object-free synchronization elements
    volatile struct _cpx_datatype_tls_context* prev;
    volatile struct _cpx_datatype_tls_context* next;

} cpx_datatype_tls_context;

#define CPX_CONST_COLOAD_MAX_CONTEXTS 32

typedef struct _cpx_datatype_coload_context
{
    /// backward compatibility
    int version;

    /// file name pinter (points to fname)
    char* file;

    /// source file line number
    int line;

    /// maximum available GPU/other co-loading contexts
    int max_contexts;

    /// co-loading contexts
    void* context[CPX_CONST_COLOAD_MAX_CONTEXTS];

    /// function capabilities (items per work-group or thread block)
    int caps[CPX_CONST_COLOAD_MAX_CONTEXTS];

    /// source file name storage
    char fname[MAX_PATH];

} cpx_datatype_coload_context;

#pragma pack(pop)

/// ---------------------------------------------------------------------------
/// Internal Runtime Data
/// ---------------------------------------------------------------------------

#define CPX_CONST_STATE_PARALLEL        0
#define CPX_CONST_STATE_SERIAL_SYNC     1
#define CPX_CONST_STATE_SERIAL_DEFAULT  2

#define CPX_CONST_MARK_INPUT  1
#define CPX_CONST_MARK_OUTPUT 2
#define CPX_CONST_MARK_DUAL   3
#define CPX_CONST_MARK_CONST  4
#define CPX_CONST_MARK_RANGE  5
#define CPX_CONST_MARK_ERROR  6
#define CPX_CONST_MARK_CACHE  7
#define CPX_CONST_MARK_WARP   8

#define CPX_CONST_MARK_END    0
#define CPX_CONST_MARK_LAST  -1

/// ---------------------------------------------------------------------------
/// Internal Runtime Functions and Macros
/// ---------------------------------------------------------------------------

/// defines the number of function arguments to share by parallel tasks
#define CPX_MACRO_SHARED_ARGS 8

/// sets up a mode to detect returns and gotos out of parallel regions
///#define CPX_MACRO_DEBUG_MODE

/// sets up a work balancing mode that is suitable for deep recursion
#define CPX_MACRO_STACK_FRIENDLY_MODE

#ifndef CPX_MACRO_OMPCOMP

/// enables the use of PAUSE instruction in barriers
#define CPX_MACRO_YIELD_MODE

#endif

#ifdef CPX_MACRO_YIELD_MODE

#define CPX_MACRO_YIELD YieldProcessor

#else

#define CPX_MACRO_YIELD SwitchToThread

#endif

/// checks if a macro was invoked with zero (false) or one argument (true)
#define CPX_MACRO_ONE_ARG(...) cpx_intrinsic_one_arg(0, __VA_ARGS__, 1, 0)
#define CPX_MACRO_ONE_PTR(...) cpx_intrinsic_one_ptr                         \
                                (0, __VA_ARGS__, (void*)1, (void*)0)

__inline int cpx_intrinsic_one_arg(size_t _0, size_t _1, size_t _2, ...)
{
    (void)_0, (void)_1;
    return (int)_2;
}

__inline int cpx_intrinsic_one_ptr(size_t _0, void* _1, void* _2, ...)
{
    (void)_0, (void)_1;
    return (int)(size_t)_2;
}

/// fetches an optional argument
__inline size_t cpx_intrinsic_get_arg(int argc, size_t arg, ...)
{
    return arg;
}

/// compiler inlining fix
#ifdef __INTEL_COMPILER
__declspec(noinline)
#else
__inline
#endif

/// fetches an optional untyped argument
size_t cpx_intrinsic_get_vararg(int argc, /*size_t arg*/...)
{
    return *((size_t*)&argc + 1);
}

#ifdef CPX_MACRO_DEBUG_MODE

/// obtains the current execution address
__declspec(noinline) static size_t cpx_intrinsic_get_instruction_pointer()
{
    return (size_t)_ReturnAddress();
}

#endif

/// parallel operator support functions
extern void CPX_EXPORTED_API cpx_exported_parallel
                       (int x, cpx_datatype_region_context* rgnctx);
extern CPX_EXPORTED_API void* cpx_exported_set_exectx
                       (cpx_datatype_execution_context* exectx, void* frame);
extern void CPX_EXPORTED_API cpx_exported_swap_exectx
                       (cpx_datatype_execution_context* exectx);
extern int CPX_EXPORTED_API cpx_exported_pick_job();
extern int CPX_EXPORTED_API cpx_exported_pick_nested_job
                       (cpx_datatype_region_context* currctx);


/// serial operator support functions
extern void CPX_EXPORTED_API cpx_exported_sync(size_t addr);
extern void CPX_EXPORTED_API cpx_exported_unsync(size_t addr);
extern int CPX_EXPORTED_API cpx_exported_trysync(size_t addr);
extern void CPX_EXPORTED_API cpx_exported_sync_default();
extern void CPX_EXPORTED_API cpx_exported_unsync_default();

/// parallel task identification support
extern CPX_EXPORTED_API void* cpx_exported_pid();
extern int CPX_EXPORTED_API cpx_exported_pix();
extern int CPX_EXPORTED_API cpx_exported_pno();

/// parallel task synchronization support
extern void CPX_EXPORTED_API cpx_exported_wake(void* pid);
/// returns true if timed out, and false otherwise
extern int CPX_EXPORTED_API cpx_exported_wait
                       (int argc, /*int milliseconds*/...);

/// data decomposition support
extern int CPX_EXPORTED_API cpx_exported_range_start(int size);
extern int CPX_EXPORTED_API cpx_exported_range_end(int size);

/// co-load support function
extern int CPX_EXPORTED_API cpx_exported_coload
                       (cpx_datatype_region_context* rgnctx, 
                        cpx_datatype_coload_context* ctx, ...);

/// ===========================================================================
/// C= Language Syntax Part
/// ===========================================================================

/// ---------------------------------------------------------------------------
/// Language-Defined Operators
/// ---------------------------------------------------------------------------

#ifdef CPX_MACRO_DEBUG_MODE

/// parallel operator
#define CPX_KEYWORD_PARALLEL(...)                                            \
{                                                                            \
    int cpx_local_region_state = CPX_CONST_STATE_PARALLEL;                   \
    size_t cpx_local_opt_arg = cpx_intrinsic_get_arg(1,__VA_ARGS__, 0);      \
    cpx_datatype_region_context cpx_local_rgnctx;                            \
    cpx_datatype_region_context* cpx_local_rgnctx_ptr;                       \
    (void)cpx_local_region_state;                                            \
                                                                             \
    if (!(cpx_local_rgnctx_ptr = (cpx_datatype_region_context*)              \
        cpx_exported_set_exectx(&cpx_local_rgnctx.exectx,                    \
        (size_t*)_AddressOfReturnAddress() + sizeof(void*) *                 \
        (CPX_MACRO_SHARED_ARGS + 1))))                                       \
    {                                                                        \
        cpx_local_rgnctx_ptr = &cpx_local_rgnctx;                            \
        cpx_exported_parallel((int)cpx_local_opt_arg, &cpx_local_rgnctx);    \
    }                                                                        \
    else                                                                     \
    {                                                                        \
        cpx_local_rgnctx_ptr->curr->rgnctx = cpx_local_rgnctx_ptr;           \
        ((size_t*)_AddressOfReturnAddress())[CPX_MACRO_SHARED_ARGS] =        \
            cpx_intrinsic_get_instruction_pointer();                         \
        *(size_t*)_AddressOfReturnAddress() = 0;                             \
        _InterlockedDecrement(&cpx_local_rgnctx_ptr->head_rgn->busy);        \
    }
#else

/// parallel operator
#define CPX_KEYWORD_PARALLEL(...)                                            \
{                                                                            \
    int cpx_local_region_state = CPX_CONST_STATE_PARALLEL;                   \
    size_t cpx_local_opt_arg = cpx_intrinsic_get_arg(1,__VA_ARGS__, 0);      \
    cpx_datatype_region_context cpx_local_rgnctx;                            \
    cpx_datatype_region_context* cpx_local_rgnctx_ptr;                       \
    (void)cpx_local_region_state;                                            \
                                                                             \
    if ((cpx_local_rgnctx_ptr = (cpx_datatype_region_context*)               \
        cpx_exported_set_exectx(&cpx_local_rgnctx.exectx,                    \
        (size_t*)_AddressOfReturnAddress() + sizeof(void*) *                 \
        (CPX_MACRO_SHARED_ARGS))) == 0)                                      \
    {                                                                        \
        cpx_local_rgnctx_ptr = &cpx_local_rgnctx;                            \
        cpx_exported_parallel((int)cpx_local_opt_arg, &cpx_local_rgnctx);    \
    }                                                                        \
    else                                                                     \
    {                                                                        \
        cpx_local_rgnctx_ptr->curr->rgnctx = cpx_local_rgnctx_ptr;           \
        _InterlockedDecrement(&cpx_local_rgnctx_ptr->head_rgn->busy);        \
    }

#endif

#ifdef CPX_MACRO_STACK_FRIENDLY_MODE

/// barrier operator (default for parallel regions)
#define CPX_KEYWORD_BARRIER()                                                \
    ;}                                                                       \
    cpx_local_opt_arg =                                                      \
        _InterlockedDecrement(&cpx_local_rgnctx_ptr->head_rgn->bcount);      \
                                                                             \
    if (!cpx_local_rgnctx_ptr->pix)                                          \
    {                                                                        \
        if (cpx_local_rgnctx_ptr->type == cpx_const_task)                    \
        {                                                                    \
            for (;;)                                                         \
            {                                                                \
                if (cpx_local_rgnctx_ptr->bcount &&                          \
                    (cpx_local_rgnctx_ptr->lcount ||                         \
                    !cpx_local_rgnctx_ptr->prev_nest))                       \
                {                                                            \
                    if (!cpx_exported_pick_nested_job(cpx_local_rgnctx_ptr)) \
                    {                                                        \
                        if (cpx_local_rgnctx_ptr->bcount)                    \
                        {                                                    \
                            CPX_MACRO_YIELD();                               \
                        }                                                    \
                        else                                                 \
                        {                                                    \
                            break;                                           \
                        }                                                    \
                    }                                                        \
                }                                                            \
                else if (!cpx_local_rgnctx_ptr->bcount)                      \
                {                                                            \
                    break;                                                   \
                }                                                            \
                else                                                         \
                {                                                            \
                    CPX_MACRO_YIELD();                                       \
                }                                                            \
            }                                                                \
        }                                                                    \
        else                                                                 \
        {                                                                    \
            while (cpx_local_rgnctx_ptr->bcount)                             \
            {                                                                \
               WaitForSingleObject                                           \
                    (cpx_local_rgnctx_ptr->head->wakesig, INFINITE);         \
            }                                                                \
        }                                                                    \
        cpx_local_rgnctx_ptr->curr->rgnctx = cpx_local_rgnctx_ptr->prev_nest;\
    }                                                                        \
    else                                                                     \
    {                                                                        \
        if (!cpx_local_opt_arg)                                              \
        {                                                                    \
            SetEvent(cpx_local_rgnctx_ptr->head->wakesig);                   \
        }                                                                    \
        cpx_local_rgnctx_ptr->curr->rgnctx = cpx_local_rgnctx_ptr->prev_nest;\
        cpx_exported_swap_exectx(cpx_local_rgnctx_ptr->idle);                \
    }

#else

/// barrier operator (default for parallel regions)
#define CPX_KEYWORD_BARRIER()                                                \
    ;}                                                                       \
    cpx_local_opt_arg =                                                      \
        _InterlockedDecrement(&cpx_local_rgnctx_ptr->head_rgn->bcount);      \
                                                                             \
    if (!cpx_local_rgnctx_ptr->pix)                                          \
    {                                                                        \
        if (cpx_local_rgnctx_ptr->type == cpx_const_task)                    \
        {                                                                    \
            for (;;)                                                         \
            {                                                                \
                if (!cpx_exported_pick_nested_job(cpx_local_rgnctx_ptr))     \
                {                                                            \
                    if (cpx_local_rgnctx_ptr->bcount)                        \
                    {                                                        \
                        CPX_MACRO_YIELD();                                   \
                    }                                                        \
                    else                                                     \
                    {                                                        \
                        break;                                               \
                    }                                                        \
                }                                                            \
            }                                                                \
        }                                                                    \
        else                                                                 \
        {                                                                    \
            while (cpx_local_rgnctx_ptr->bcount)                             \
            {                                                                \
               WaitForSingleObject                                           \
                    (cpx_local_rgnctx_ptr->head->wakesig, INFINITE);         \
            }                                                                \
        }                                                                    \
        cpx_local_rgnctx_ptr->curr->rgnctx = cpx_local_rgnctx_ptr->prev_nest;\
    }                                                                        \
    else                                                                     \
    {                                                                        \
        if (!cpx_local_opt_arg)                                              \
        {                                                                    \
            SetEvent(cpx_local_rgnctx_ptr->head->wakesig);                   \
        }                                                                    \
        cpx_local_rgnctx_ptr->curr->rgnctx = cpx_local_rgnctx_ptr->prev_nest;\
        cpx_exported_swap_exectx(cpx_local_rgnctx_ptr->idle);                \
    }

#endif

/// barrier skipping operator (exits parallel and serial regions)
#define CPX_KEYWORD_BREAK()                                                  \
    ;}                                                                       \
    if (cpx_local_region_state == CPX_CONST_STATE_PARALLEL)                  \
    {                                                                        \
        if (!cpx_local_rgnctx_ptr->pix)                                      \
        {                                                                    \
            if (cpx_local_rgnctx_ptr->type == cpx_const_task)                \
            {                                                                \
                for (;;)                                                     \
                {                                                            \
                    if (!cpx_exported_pick_nested_job(cpx_local_rgnctx_ptr)) \
                    {                                                        \
                        if (cpx_local_rgnctx_ptr->busy)                      \
                        {                                                    \
                            CPX_MACRO_YIELD();                               \
                        }                                                    \
                        else                                                 \
                        {                                                    \
                            break;                                           \
                        }                                                    \
                    }                                                        \
                }                                                            \
            }                                                                \
            else                                                             \
            {                                                                \
                while (cpx_local_rgnctx_ptr->busy) SwitchToThread();         \
            }                                                                \
            cpx_local_rgnctx_ptr->head->rgnctx =                             \
                cpx_local_rgnctx_ptr->prev_nest;                             \
        }                                                                    \
        else                                                                 \
        {                                                                    \
            cpx_exported_swap_exectx(cpx_local_rgnctx_ptr->idle);            \
        }                                                                    \
    }                                                                        \
    else                                                                     \
    {                                                                        \
        (cpx_local_region_state == CPX_CONST_STATE_SERIAL_SYNC) ?            \
            cpx_exported_unsync(cpx_local_opt_arg) :                         \
            cpx_exported_unsync_default();                                   \
    }

/// serial operator
#define CPX_KEYWORD_SERIAL(...)                                              \
{                                                                            \
    int cpx_local_region_state = CPX_CONST_STATE_SERIAL_SYNC;                \
    cpx_datatype_region_context* cpx_local_rgnctx_ptr = 0;                   \
    size_t cpx_local_opt_arg = cpx_intrinsic_get_vararg(1,__VA_ARGS__, 0);   \
                                                                             \
    if (CPX_MACRO_ONE_PTR(__VA_ARGS__))                                      \
    {                                                                        \
        cpx_exported_sync(cpx_local_opt_arg);                                \
    }                                                                        \
    else                                                                     \
    {                                                                        \
        cpx_local_region_state = CPX_CONST_STATE_SERIAL_DEFAULT;             \
        cpx_exported_sync_default();                                         \
    }

/// serial operator with try-lock semantics

#define CPX_MACRO_TRY_serial(a) cpx_exported_trysync((size_t)a)
#define CPX_MACRO_GET_serial(a) cpx_intrinsic_get_vararg(1, (size_t)a)

#define try(a)                                                               \
                                                                             \
    if (CPX_MACRO_TRY_##a)                                                   \
    {                                                                        \
        int cpx_local_region_state = CPX_CONST_STATE_SERIAL_SYNC;            \
        cpx_datatype_region_context* cpx_local_rgnctx_ptr = 0;               \
        size_t cpx_local_opt_arg = CPX_MACRO_GET_##a;

/// coload operator
#define CPX_KEYWORD_COLOAD(...)                                              \
                                                                             \
    int cpx_local_coload_continue = 1;                                       \
                                                                             \
    if(!pix())                                                               \
    {                                                                        \
        static cpx_datatype_coload_context ctx =                             \
            {0, &ctx.fname[0], __LINE__, CPX_CONST_COLOAD_MAX_CONTEXTS,      \
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                 \
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                 \
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                 \
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, {__FILE__} };   \
                                                                             \
        cpx_local_coload_continue = cpx_exported_coload                      \
            (0, &ctx, __VA_ARGS__, 0);                                       \
    }                                                                        \
    if(cpx_local_coload_continue)

/// ---------------------------------------------------------------------------
/// Language-Defined Data Types and Modifiers
/// ---------------------------------------------------------------------------

/// modifier to ensure shared variables are not kept in registers
#define CPX_KEYWORD_SHARED volatile

/// modifiers to specify co-load data dependencies and optimization parameters

#define CPX_KEYWORD_COLOAD_INPUT(type, var, ...)                             \
        CPX_CONST_MARK_INPUT, #type, var, sizeof(var),                       \
        __VA_ARGS__, CPX_CONST_MARK_LAST

#define CPX_KEYWORD_COLOAD_OUTPUT(type, var, ...)                            \
        CPX_CONST_MARK_OUTPUT, #type, var, sizeof(var),                      \
        __VA_ARGS__, CPX_CONST_MARK_LAST

#define CPX_KEYWORD_COLOAD_DUAL(type, var, ...)                              \
        CPX_CONST_MARK_DUAL, #type, var, sizeof(var),                        \
        __VA_ARGS__, CPX_CONST_MARK_LAST

#define CPX_KEYWORD_COLOAD_CONST(type, var)                                  \
        CPX_CONST_MARK_CONST, #type, &var, sizeof(var)

#define CPX_KEYWORD_COLOAD_RANGE(range) CPX_CONST_MARK_RANGE, range
#define CPX_KEYWORD_COLOAD_ERROR(error) CPX_CONST_MARK_ERROR, error
#define CPX_KEYWORD_COLOAD_CACHE(cache) CPX_CONST_MARK_CACHE, cache
#define CPX_KEYWORD_COLOAD_WARP(warp)   CPX_CONST_MARK_WARP,  warp

/// ---------------------------------------------------------------------------
/// Language-Defined (Built-In) Functions
/// ---------------------------------------------------------------------------

/// system-unique parallel task ID
#define CPX_KEYWORD_PID() cpx_exported_pid()
/// parallel region-unique task index
#define CPX_KEYWORD_PIX() cpx_exported_pix()
/// the number of tasks in the current parallel region
#define CPX_KEYWORD_PNO() cpx_exported_pno()

/// signaling operator to pass control to a task
#define CPX_KEYWORD_WAKE cpx_exported_wake
/// control suspension operator to wait until the control is returned
#define CPX_KEYWORD_WAIT(...) cpx_exported_wait                              \
                               (CPX_MACRO_ONE_ARG(__VA_ARGS__), __VA_ARGS__)

/// ---------------------------------------------------------------------------
/// Language-Defined Auxiliary Macros
/// ---------------------------------------------------------------------------

/// data-decomposition macros
#define CPX_MACRO_RANGE_START(size) cpx_exported_range_start(size)
#define CPX_MACRO_RANGE_END(size) cpx_exported_range_end(size)

/// ===========================================================================
/// C= End of Header
/// ===========================================================================

#ifdef __cplusplus
}
#endif

#endif ///_CPX_HEADER_H_
