//
//  File  : nbody.cpp
//  (c)HOOPOE ProGroup 2013
//

//
// C= Programming Sample
//

/// Copyright string
char* copyright = "(c)HOOPOE ProGroup 2013";

#include <stdio.h>
#include <C=.h>

#include <stdlib.h>
#include <math.h>

#include <intrin.h>
#pragma intrinsic(__rdtsc)

__int64 time;

/// data

const int N=16384;///4096

float pos[N * 4];
float vel[N * 4];

float force[N * 3];

int bodies = N;
 
float soften = .00125f;
float pscale = 1.54f;
float vscale = 8.0f;
float damp = 0.995f;
float step = 0.016f;

/// auxilliary functions

float dot(float v0[3], float v1[3])
{
    return v0[0] * v1[0] + v0[1] * v1[1] + v0[2] * v1[2];
}

/// initialize N-bodies

void init_bodies(float* pos, float* vel, float pscale, float vscale, int bodies)
{
    float scale = pscale * bodies / 1024.f;

    if(scale < 1.0f)
    {
        scale = pscale;
    }
    vscale *= scale;

    int p = 0, v = 0;

    for(int i = 0; i < bodies;) 
    {
        float point[3];

        point[0] = rand() / (float)RAND_MAX * 2 - 1;
        point[1] = rand() / (float)RAND_MAX * 2 - 1;
        point[2] = rand() / (float)RAND_MAX * 2 - 1;

        float len = dot(point, point);
        
        if (len > 1)
        {
            continue;
        }

        pos[p++] = point[0] * scale; /// X
        pos[p++] = point[1] * scale; /// Y
        pos[p++] = point[2] * scale; /// Z
        pos[p++] = 1.0f; // mass
        vel[v++] = point[0] * vscale; /// X-velocity
        vel[v++] = point[1] * vscale; /// Y-velocity
        vel[v++] = point[2] * vscale; /// Z-velocity
        v++;

        i++;
    }
}

/// define rsqrtf for CPU to match CUDA's definition
/// change it to rsqrt to match OpenCL's definition
#define rsqrtf(x) (1.0f / sqrtf(x))

/// compute N-bodies

void compute_bodies(float step)
{
    /// convert the time step to float (it is passed as double by the compiler)
    float delta = (float)step;

    /// use parallel(1) to send everything to GPU if available
    parallel(1)
    {
        /// set up memory local to GPU threads 
        /// we intend to use 256 threads per execution block
        float cache[256 * 4];
        
        /// co-load all GPUs (and describe all dependencies)
        coload(
               /// input positions
               cpx_in (float*, pos,   bodies * sizeof(float) * 4),
               /// output forces
               cpx_out(float*, force, bodies * sizeof(float) * 3),

               /// input constants
               cpx_const(int, bodies), 
               cpx_const(float, damp), 
               cpx_const(float, soften),

               /// specify the size of the memory buffer shared by GPU threads 
               /// running on the same multiprocessor
               cpx_cache(sizeof(cache)),
               /// requested number of threads to run on a multiprocessor 
               /// (threads per thread block)
               cpx_warp(256),

               /// the actual number of iterations to perform
               cpx_range(bodies))

               /// NOTE: ((cpx_range / pno()) % cpx_warp) must be 0
        {
            /// loop over bodies (use built-in decomposition macros)
            for (int i = CPX_RANGE_START(bodies), 
                   end = CPX_RANGE_END(bodies) ; i < end; i++)
            {
                /// compute array indices
                int idx = 4 * i;
                int fdx = 3 * i;

                /// temporary accumulator
                float acc[3] = {0, 0, 0};

                /// work in 256-body blocks
                for(int j = 0; j < bodies; j += 256) 
                {
                    /// fill in a shared memory buffer 
                    /// (256 interations on CPU, 1 iteration on GPU)
                    for(int k = i; k < ((end - i) < 256 ? end : (i + 256)); k++)
                    {
                        int jdx = 4 * (j + (k % 256));
                        int kdx = 4 * (k % 256);

                        /// copy k-th element from the current 256-body block
                        /// to k-th element in the shared buffer
                        cache[kdx + 0] = pos[jdx + 0];
                        cache[kdx + 1] = pos[jdx + 1];
                        cache[kdx + 2] = pos[jdx + 2];
                        cache[kdx + 3] = pos[jdx + 3];
                    }
                    /// wait for all GPU threads to complete the operation
                    cpx_sync();

                    /// compute i-th body interactions with all bodies
                    /// within the current 256-body block
                    for(int l = 0; l < 256; l++)
                    {
                        int jdx = 4 * l;
        	
                        float r[3];

                        /// compute vector deltas
                        r[0] = cache[jdx + 0] - pos[idx + 0];
                        r[1] = cache[jdx + 1] - pos[idx + 1];
                        r[2] = cache[jdx + 2] - pos[idx + 2];

                        /// compute distance
                        float dist = r[0] * r[0] + r[1] * r[1] + r[2] * r[2];
                        dist += soften;

                        /// 1 / dist^(3/2)
                        float inv = rsqrtf(dist);
                        float cube =  inv * inv * inv;

                        /// compute force
                        float s = cache[jdx + 3] * cube;

                        /// superpose forces
                        acc[0] += r[0] * s;
                        acc[1] += r[1] * s;
                        acc[2] += r[2] * s;
                    }
                    /// wait for all GPU threads to complete their operation
                    cpx_sync();
                }
                /// store forces
                force[fdx + 0] = acc[0];
                force[fdx + 1] = acc[1];
                force[fdx + 2] = acc[2];
            }
        }
        barrier();
    }

    /// use parallel(1) to send everything to GPU if available
    parallel(1)
    {
        /// co-load all GPUs (and describe all dependencies)
        coload(
               /// input forces
               cpx_in   (float*, force, bodies * sizeof(float) * 3),
               /// input and output velocities
               cpx_inout(float*, vel,   bodies * sizeof(float) * 4),
                /// input and output positions
               cpx_inout(float*, pos,   bodies * sizeof(float) * 4),

               /// input constants
               cpx_const(float, damp),
               cpx_const(float, delta),
                    
               /// the actual number of iterations to perform
               cpx_range(bodies))
        {
            /// loop over bodies (use built-in decomposition macros)
            for (int i = CPX_RANGE_START(bodies), 
                   end = CPX_RANGE_END(bodies) ; i < end; ++i)
            {
                /// position, velocity and force indices
                int idx = 4 * i;
                int fdx = 3 * i;

                /// read positions, velocities, forces, masses 
                /// into local variables
                float p[3], v[3], f[3];

                p[0] = pos[idx + 0];
                p[1] = pos[idx + 1];
                p[2] = pos[idx + 2];

                float mass = pos[idx + 3];
                
                v[0] = vel[idx + 0];
                v[1] = vel[idx + 1];
                v[2] = vel[idx + 2];

                f[0] = force[fdx + 0];
                f[1] = force[fdx + 1];
                f[2] = force[fdx + 2];

                /// compute and superpose new velocities and positions

                // acceleration = force / mass; 
                // new velocity = old velocity + acceleration * delta time
                v[0] += (f[0] * mass) * delta;
                v[1] += (f[1] * mass) * delta;
                v[2] += (f[2] * mass) * delta;

                v[0] *= damp;
                v[1] *= damp;
                v[2] *= damp;

                // new position = old position + velocity * delta time
                p[0] += v[0] * delta;
                p[1] += v[1] * delta;
                p[2] += v[2] * delta;

                /// store positions
                pos[idx + 0] = p[0];
                pos[idx + 1] = p[1];
                pos[idx + 2] = p[2];

                /// store velocities
                vel[idx + 0] = v[0];
                vel[idx + 1] = v[1];
                vel[idx + 2] = v[2];
            }
        }
        barrier();
    }
}

void main()
{
    /// prime the coload functionality 
    parallel(1)
    {
        /// use empty coload operator to trigger lazy initialization
        coload(){}
        break();
    }

    /// initialize body system
    init_bodies(pos, vel, pscale, vscale, bodies);

    time = __rdtsc();

    /// update the system
    compute_bodies(step);

    time = __rdtsc() - time;

    printf("Time: %I64d\n\n", time);

    /// print the results
    for (int i = 0; i < bodies; i++)
    {
        printf("H[%d]=%f\n", i, pos[i]);
    }
}
